#include <iostream>

#include "solver_cxx.h"
#include "util/conversions.h"

#include "mongocxx/instance.hpp"
#include "mongocxx/client.hpp"
#include "mongocxx/database.hpp"
#include "mongocxx/collection.hpp"
#include "mongocxx/cursor.hpp"

#include "INIReader.h"

#include "util/wmis_qubo.h"

std::string getDbConfigFilePath();
std::string getScopeName();
std::string getDBUrl();
mongocxx::client connectToDatabase();
void print(const qbscxx::SOLUTION& solution, const satlab::wmis_qubo& qubo);

int main(int argc, char** argv)
{
    mongocxx::instance inst{};
    
    mongocxx::client conn = connectToDatabase();
    
    mongocxx::database db = conn["experiments"];
    
    mongocxx::collection wmis_qubos = db["wmis_qubos"];
    
    
    std::cout << "runQbsolv" << std::endl;
    
    satlab::wmis_qubo wmis_qubo;
    
    for(auto& el : wmis_qubos.find({}))
    {   
        satlab::types::convert(wmis_qubo, el);
        
        qbscxx::Solver solver(wmis_qubo.get_qubo());
        
        solver.set_goal(qbscxx::Solver::GOAL::MAXIMUM);
        solver.set_verbosity(-1);
        
        solver.solve();
        
        print(solver.get_solution(), wmis_qubo);
        
        std::cout << std::endl;
        
    }
    
    return 0;
}


mongocxx::client connectToDatabase()
{
    mongocxx::options::ssl sslOptions;
    sslOptions.allow_invalid_certificates(true);
     
	mongocxx::options::client clientOptions{};
	clientOptions.ssl_opts(sslOptions);
    
	return mongocxx::client{mongocxx::uri{getDBUrl()}, clientOptions};
}

std::string getDBUrl()
{
	INIReader iniReader(getDbConfigFilePath());

	std::string connUrl("mongodb://");
	connUrl.append(iniReader.Get("CONNECTION", "user", "NA") + ":");
	connUrl.append(iniReader.Get("CONNECTION", "pw", "NA") + "@");
	connUrl.append(iniReader.Get("CONNECTION", "url", "NA") + ":");
	connUrl.append(iniReader.Get("CONNECTION", "port", "NA") + "/");
	connUrl.append(iniReader.Get("CONNECTION", "database", "NA"));
	connUrl.append("?ssl=true");

	return connUrl;
}

std::string getDbConfigFilePath()
{
	std::cout << "Database config file: ";

	std::string path;

	std::cin >> path;

	return path;
}

std::string getScopeName()
{
	std::cout << "Experiment scope: ";

	std::string scope;

	std::cin >> scope;

	return scope;
}

void print(const qbscxx::SOLUTION& solution, const satlab::wmis_qubo& qubo)
{
    qbscxx::ENERGY_T resulution_energy = 0;
        for(int val : solution)
        {
            resulution_energy += val;
            
            std::cout << val;
        }
        
        std::cout << " " << resulution_energy << std::endl;
        
        std::vector<bool> model;
        int conflicts = 0;
        for(int i = 0; i < solution.size(); i++)
        {
            auto node = qubo.node_at(i);
            int lit = std::get<1>(node);
            size_t var = std::abs(lit);
            
            bool newVar = false;
            if(model.size() < var)
            {
                newVar = true;
                model.resize(var);
            }
            
            bool binding;
            if(solution[i] == 0)
            {
                binding = lit < 0 ? true : false;
            }
            else
            {
                binding = lit < 0 ? false : true;
            }
            
            if(!newVar && binding != model[var - 1])
            {
                conflicts++;
            }
            
            model[var - 1] = binding;
        }
        
        
        for(int i=0; i < model.size(); i++)
        {
            int var = i + 1;
            std::cout << (model[i] ? var : -1 * var) << " ";
        }
        
        std::cout << std::endl;
}


