#include "util/wmis_qubo.h"
#include <algorithm>

namespace satlab
{
    
void wmis_qubo::add(const COUPLER& coupler)
{
    const NODE& node1 = std::get<0>(coupler);
    const NODE& node2 = std::get<1>(coupler);
    qbscxx::ENERGY_T energy = std::get<2>(coupler);
    
    size_t index_node1 = add(node1);
    size_t index_node2 = add(node2);
    
    fit_qubo(std::max(index_node1, index_node2));
    
    qubo[index_node1][index_node2] = energy;
    qubo[index_node2][index_node1] = energy;
}

size_t wmis_qubo::add(const NODE& node)
{
    size_t index;
    
    if(node_to_index.find(node) == node_to_index.end())
    {
        index_to_node.push_back(node);
        index = index_to_node.size() - 1;
        
        node_to_index.emplace(node, index);
    }
    else
    {
        index = index_of(node);
    }
    
    return index;
}

void wmis_qubo::fit_qubo(const size_t node_index)
{
    size_t required_size = node_index + 1;
    
    if(qubo.size() >= required_size)
    {
        return;
    }
    
    size_t old_size = qubo.size();
    
    qubo.resize(required_size);
    for(size_t i=0; i < required_size; i++)
    {
        qubo[i].resize(required_size);
        
        std::fill(qubo[i].begin() + old_size, qubo[i].end(), 0.0);
    }
}


size_t wmis_qubo::index_of(const NODE& node) const
{
    return node_to_index.at(node);
}

const satlab::wmis_qubo::NODE& wmis_qubo::node_at(size_t index) const
{
    return index_to_node.at(index);
}

const qbscxx::QUBO& wmis_qubo::get_qubo() const
{
    return qubo;
}



}
