#ifndef WMIS_QUBO_H
#define WMIS_QUBO_H

#include "types.h"
#include <map>
#include <tuple>
#include <vector>

namespace satlab
{
    
/**
 * @todo write docs
 */
class wmis_qubo
{
public:
    using NODE = std::tuple<int, int>;
    using COUPLER = std::tuple<NODE, NODE, qbscxx::ENERGY_T>;
    
    using INDEX_TO_NODE = std::vector<NODE>;
    using NODE_TO_INDEX = std::map<NODE, size_t>;
    
    wmis_qubo() = default;
    virtual ~wmis_qubo() = default;
    
    wmis_qubo(const wmis_qubo& other) = default;
    wmis_qubo(wmis_qubo&& other) = default;
    
    wmis_qubo& operator=(const wmis_qubo& other) = default;
    wmis_qubo& operator=(wmis_qubo&& other) = default;
    
    
    void add(const COUPLER& coupler);
    
    const NODE& node_at(size_t index) const;
    size_t index_of(const NODE& node) const;
    
    const qbscxx::QUBO& get_qubo() const;
private:
    size_t add(const NODE& node);
    void fit_qubo(const size_t node_index);
    
    qbscxx::QUBO qubo;
    INDEX_TO_NODE index_to_node;
    NODE_TO_INDEX node_to_index;
};

}

#endif // WMIS_QUBO_H

